/* Copyright (C) 1999, Vsoft Ltd. All Rights Reserved.
 *
 *   Usage Restrictions
 *
 * License is granted to copy, to use, and to make and to use derivative
 * works for research and evaluation purposes only.
 *
 *   Disclaimer of Warranty
 *
 * These software programs are available to the user without any license
 * fee or royalty on an "as is" basis.  Vsoft Ltd. disclaims any and
 * all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any 
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and
 * user's customers, employees, agents, transferees, successors, and
 * assigns.
 *
 * Comments & Questions:
 * mailto:noam@vsoft.com  . Please put "TestPoint" in the subject line.
 */

#ifndef _ITestPoint_H
#define _ITestPoint_H

#include <objbase.h>


// {E2B1B461-406D-11d3-AEF9-000000000000}
static const GUID IID_TestPointManager = 
{ 0xe2b1b461, 0x406d, 0x11d3, { 0xae, 0xf9, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0 } };

// {E2B1B462-406D-11d3-AEF9-000000000000}
static const GUID IID_ITestPoint = 
{ 0xe2b1b462, 0x406d, 0x11d3, { 0xae, 0xf9, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0 } };

// {E2B1B463-406D-11d3-AEF9-000000000000}
static const GUID CLSID_TestPointManager = 
{ 0xe2b1b463, 0x406d, 0x11d3, { 0xae, 0xf9, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0 } };


/*
Interface to test points in our lovely software.
Testpoints are data probes activated by the administrator at run time. 
A pieace of software can check for the activity of its testpoint
(assuming a testpoint was built into it). 
If the testpoint is active, an interface is supplied for dumping the data.

The ITestPointManager interface itself does not facilitate dumping data.
The user ( i.e. the module trying to dump data to the testpoint) has to get
a handler; this handler ( a COM object ) exposes the required interface.

This architecture isolate the specific TP handlers from the TP manager, allowing 
for dynamic growth of the TP system.

In the future, a 'third party' observer might wish to interact with specific handlers,
e.g show the user wich TP are active, what is written to the TP, turn the TP on/off etc.
This communication shall be conducted with the TP manager using interfaces not yet defined.


Sample usage:
MyClass::WriteIt(char* buff, int len)
{
	IDataSink2 * pOutput;
	if(NULL != pTestPointManger // ITestPoint * pTestPointManger
	   && ( S_OK == pTestPointManger->IsActive(VcrCoreTp1) ) // arg is a GUID
	{
		IDataSink2 pSink;
		if(SUCCEEDED(pTestPoint->QueryInterface(IID_DATASINK,(void**) pSink))
		{
			// connect a T sink
			// ...
			// and use the T.
			pOutput = pSink
		}
		// handle the error - you can ignore it silently
		pOutput = m_pNormalFlowSink; // silent ignore
	}
	else
	{
		pOutput = m_pNormalFlowSink;
	}

  // other stuff

  // release 
	pSink->Release();
}

IMPLEMENTATION ( informative )
This interface is implemented by a COM object ( possibly an outproc server) called the 
TP manager. ( don't confuse with transaction processor...)
When the manager is instantiated, it queries some database ( e.g. the registry) for the
known test points and their parameters. 
All TP must be known to the manager before usage.

Example of parameters:
ID:
{F2EB56C0-403A-11d3-9DBD-0060089927C2}

Handler:
{F2EB56C1-403A-11d3-9DBD-0060089927C2}

Name:
bitstream at input to splitter

Active:
0x0001 ( true )

File:
C:\temp\splitIn.mpg


*/
interface ITestPointManager : IUnknown
{
	/* check if a specific TP is active. 
	 Activity state is initialy set by the 'outside world' - in the registry maybe,
	 and can be changed by calling SetActive()

	 RETURN:
	 S_OK = the TP is active (i.e. you should write to it)
	 S_FALSE = the TP should not be used now. 
	 */
	virtual HRESULT IsActive( const IN GUID & idOfTestPoint) = 0;


	/* Get the name of the interface into the buffer supplied by the caller.
	 if the name is longer than the buffer, it is silently truncated.
	
	 RETURN:
	 S_OK
	 E_INVALIDARG if the TP id is not known to the TP manager
	 E_NULLPTR if pNameOfTp == NULL
	 */
	virtual HRESULT GetNameOfTp( const IN GUID & idOfTestPoint, // ID of the TP
			IN	 UINT bufferSize,		// maximum size of buffer 
			OUT  char* pNameOfTp		// name of the TP will be placed here
			) = 0;


	/* Set the state of the TP activity to ON / OFF.
	 The new state is valid until changed or the testPoint manager is unloaded.
	 ( the testpoint manger is the COM object implementing this interface
	 
	 RETURN:
	 S_OK if everything is well,
	 E_INVALIDARG if the TP id is not known to the TP manager
	 */
	virtual HRESULT SetActive( const IN GUID & idOfTestPoint , IN bool isActive) = 0;

	/*
	 Get an interface to a handler of the required TP
	 The handler is known by its CLSID or ProgID as appears in
	 the values for the testpoint. In there is a CLSID, it overides
	 the ProgID.
	
	 RETRUN:
	 E_INVALIDARG if the TP id is not known to the TP manager,
	 other errors as well.
	 */
	virtual HRESULT QueryInterfaceForTp(const IN GUID & idOfTestPoint , 
										OUT IUnknown ** pUnknown) =0;

};


/*
	A handler of a test point (TP) has to implement the ITestPoint.
	This interface is used by the test point manager to pass identifying information
	of the client. i.e. when TP1 has to be handled ( written to), the handler of TP1
	is instantiated and passed the ID of TP1 and location of parameters for it.
	This enables a handler to server more than one TP ( e.g. IDataSink can be used 
	in multiple places in the client object, in many threads and processes.

	It is the responsiblity of the handler to distinguish between different threads 
	and processes ( e.g. for creating uinque dump file names)
 */
interface ITestPoint : IUnknown
{
	virtual HRESULT Set(const IN GUID& idOfTestPoint,
						const IN char* parameterLocation // unique path to parameters
														// for this TP.
														// e.g. Registry path
						) =0;
};


#endif