//** CONFIDENTIAL ***** CONFIDENTIAL ***** CONFIDENTIAL ***** CONFIDENTIAL **
//   Campwood Software, Charlotte, Vermont 05445 USA
//   Copyright  1999, All Rights Reserved.
//***************************************************************************
//
// FILE:       SMDelphi.cpp	[ SMDelphi.h ]
//
// SUBSYSTEM:  Source Monitor Applicaton
//
// AUTHOR:     JF Wanner
//
// OVERVIEW:   Implements Delphi (Object Pascal) metrics language class SMDelphi
//
//***************************************************************************

#include "stdafx.h"

#include "SMLanguage.h"
#include "SMDelphi.h"
#include "SourceMetrics.h"
#include "resource.h"
#include <VPDelphiClasses.hpp>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

IMPLEMENT_SERIAL(SMDelphi, SMLanguage, 1)

// static members
CString        SMDelphi::DELPHI_NAME = _T("Delphi");
CString        SMDelphi::DELPHI_EXTENSIONS = _T("*.pas;");

int            SMDelphi::m_iParserUsers = 0;
DelphiLexClass *
               SMDelphi::m_poLexer = NULL;  // parsing objects are created and deleted by App object
SSLexTable *   SMDelphi::m_poLexTable = NULL;
SSYaccTable *  SMDelphi::m_poYaccTable = NULL;

// ***************************************************************************
// normal constructor
SMDelphi::SMDelphi() : SMLanguage()
{
   m_iCounts.SetSize(TOTAL_COUNTS);
}

// ***************************************************************************
// Class factory
SMLanguage *
SMDelphi::MakeNew()           // create a shallow copy
{
   SMDelphi *  poNewSMDelphi;
   POSITION    oPosition;
   CString     sClassName;
   CString     sClassValue;

   poNewSMDelphi = new SMDelphi();
   poNewSMDelphi->m_sSubroutineName = m_sSubroutineName;
   for( int i = 0; i < m_iCounts.GetSize(); ++i)
   {
      poNewSMDelphi->m_iCounts[i] = m_iCounts[i];
   }
   oPosition = m_oClassList.GetStartPosition(); 
   while( oPosition != NULL )
   {
      m_oClassList.GetNextAssoc(oPosition, sClassName, sClassValue);
      poNewSMDelphi->m_oClassList[sClassName] = sClassValue;
   }
   return poNewSMDelphi;
}

// ***************************************************************************
// destructor
SMDelphi::~SMDelphi()
{
}

// ***************************************************************************
// initialize global parser objects
void
SMDelphi::InitParser()
{
   if( m_iParserUsers == 0 )
   {
      m_poLexTable = new SSLexTable((unsigned int)AfxGetResourceHandle(),
                                          MAKEINTRESOURCE(IDR_DELPHI_LEXTABLE), _T("LEXTABLE"));
      m_poYaccTable = new SSYaccTable((unsigned int)AfxGetResourceHandle(),
                                          MAKEINTRESOURCE(IDR_DELPHI_YACCTABLE), _T("YACCTABLE"));
      m_poLexer = new DelphiLexClass(*SMDelphi::m_poLexTable);
   }
   ++m_iParserUsers;
}

// ***************************************************************************
// shut down parser
void
SMDelphi::CloseParser()
{
   if( m_iParserUsers > 0 )
   {
      --m_iParserUsers;
   }
   if( m_iParserUsers == 0 )
   {
      delete m_poLexTable;
      m_poLexTable = NULL;
      delete m_poYaccTable;
      m_poYaccTable = NULL;
      delete m_poLexer;
      m_poLexer = NULL;
   }
}

// ***************************************************************************
// compute and accumulate metrics for one line or one file
bool
SMDelphi::Start(MetricsType /* type */)
{
   bool  fResult = true;

   if( m_poLexer == NULL )
   {
      fResult = false;
   }
   else
   {
      m_sSubroutineName.Empty();
      for( int i = 0; i < TOTAL_COUNTS; ++i )
      {
         m_iCounts[i] = 0;       // initialize the count accumulators
      }
      m_oClassList.RemoveAll();
      m_iBeginSubroutineLine = 0;
      m_iBeginMethodLine = 0;
   }
   return fResult;
}

// ***************************************************************************
// compute and accumulate metrics for one source file
bool
SMDelphi::ProcessLines(const CString & sFileName)
{
   bool  fResult = true;         // assume success

   try
   {
      SSLexFileConsumer oSource(sFileName);
      m_poLexer->reset();
      m_poLexer->setConsumer(oSource);
      DelphiYaccClass(*m_poLexer, *m_poYaccTable, this).parse();
      SMLanguage::ReportErrors(sFileName);
   }
   catch(SSException oExcept)
   {
      m_sError.Format("%s [%s]", oExcept.text(), sFileName);
      fResult = false;
   }
   catch(...)
   {
      fResult = false;
   }
   return fResult;
}

// ***************************************************************************
// called by the parser's reduce() method when parser can't handle source code structure
void
SMDelphi::ParseError()
{
   if( AfxGetApp()->GetProfileInt(CSourceMetricsApp::OPTION_SETTINGS, CSourceMetricsApp::REPORT_ERRORS, 0) != 0 )
   {
      CString  sLineNumber;

      sLineNumber.Format("%d", m_iCounts[LINE_COUNT] + 1);
      m_asErrorLines.Add(sLineNumber);
   }
}

// ***************************************************************************
// called by the parser's reduce() method at then end of each subroutine implementation
void
SMDelphi::SetSubroutineData(const CString & sSubroutineName, int iStatementCount)
{
   m_iCounts[SUBROUTINE_STATEMENTS] += iStatementCount;
   if( (int)m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] < iStatementCount )     // save only the biggest one
   {
      m_sSubroutineName = sSubroutineName.Mid(sSubroutineName.FindOneOf(_T(" \t")));
      m_sSubroutineName.TrimLeft();
      m_iCounts[BIGGEST_SUBROUTINE_LINE] = m_iBeginSubroutineLine;
      m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] = iStatementCount;
   }
}

// ***************************************************************************
// called by the parser's reduce() method at the end of each method implementation
void
SMDelphi::SetMethodData(const CString & sClassAndMethodName, int iStatementCount)
{
   CString  sClassName;                // scratch for parsing the input string
   CString  sMethodName;               // scratch for parsing the input string
   int      iMethodOffset;
   CString  sClassValue;               // class list value for biggest method test
                                       //    value format: <line count>@<method name>
   bool     fUpdateClassList = true;   // assume list needs to be changed
   
   m_iCounts[METHOD_STATEMENTS] += iStatementCount;

   // parse the names out of the lexeme string: procedure|function <class>.<method>
   sClassName = sClassAndMethodName.Mid(sClassAndMethodName.FindOneOf(_T(" \t")));
   sClassName.TrimLeft();
   iMethodOffset = sClassName.Find('.');
   sMethodName = sClassName.Mid(iMethodOffset + 1);
   sClassName = sClassName.Left(sClassName.GetLength() - sMethodName.GetLength() - 1);
   if( m_oClassList.Lookup(sClassName, sClassValue) )   // see if the class is in the list
   {
      // check to see if biggest method is stored in the map's value
      if( _ttoi(sClassValue) > iStatementCount )
      {
         fUpdateClassList = false;     // class is there with a bigger method: do nothing
      }
   }
   if( fUpdateClassList )
   {
      sClassValue.Format("%d@%s", iStatementCount, sMethodName);
      m_oClassList[sClassName] = sClassValue;
      if( iStatementCount >= (int)m_iCounts[BIGGEST_METHOD_STATEMENTS] )
      {
         m_iCounts[BIGGEST_METHOD_LINE] = m_iBeginMethodLine;
         m_iCounts[BIGGEST_METHOD_STATEMENTS] = iStatementCount;
      }
   }
}

bool
SMDelphi::ProcessFile(const SMLanguage & oFileMetric)
{
   const SMDelphi &
            oFileData = dynamic_cast<const SMDelphi &>(oFileMetric);
   CString  sClassValue;         // class list value for biggest method test
                                 //    value format: <line count>@<method name>
   POSITION oSourcePosition;     // iterator for file's class list
   CString  sClassName;          // file's current class's name
   CString  sSourceValue;        // file's current class's value
   CString  sDestinationValue;   // value for file's class in local class list
   bool     fUpdateClassList;    // true if class list needs to be changed
   bool     fResult = true;      // assume success

   // deal with the non-sum statement counts first
   if( oFileData.m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] > m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] )
   {
      m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] = oFileData.m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS];
      m_sSubroutineName = oFileData.m_sSubroutineName;
   }
   if( oFileData.m_iCounts[BIGGEST_METHOD_STATEMENTS] > m_iCounts[BIGGEST_METHOD_STATEMENTS] )
   {
      m_iCounts[BIGGEST_METHOD_STATEMENTS] = oFileData.m_iCounts[BIGGEST_METHOD_STATEMENTS];
   }
   // then accumulate the rest of the counts
   for( int i = 0; i < TOTAL_COUNTS; ++i )
   {
      if( (i != BIGGEST_SUBROUTINE_STATEMENTS) and (i != BIGGEST_METHOD_STATEMENTS) )
      {
         m_iCounts[i] += oFileData.m_iCounts[i];   // accumulate count totals
      }
   }
   // not valid for checkpoints so reset them to flag values
   m_iCounts[DEEPEST_BLOCK_LINE] = UINT_MAX;
   m_iCounts[BIGGEST_SUBROUTINE_LINE] = UINT_MAX;
   m_iCounts[BIGGEST_METHOD_LINE] = UINT_MAX;

   // now build a list of unique classes with biggest method values
   oSourcePosition = oFileData.m_oClassList.GetStartPosition();
   while( oSourcePosition != NULL )
   {
      fUpdateClassList = true;         // assume change is needed
      oFileData.m_oClassList.GetNextAssoc(oSourcePosition, sClassName, sSourceValue);
      if( m_oClassList.Lookup(sClassName, sDestinationValue) )
      {
         if( _ttoi(sSourceValue) < _ttoi(sDestinationValue) )
         {
            fUpdateClassList = false;
         }
      }
      if( fUpdateClassList )
      {
         m_oClassList[sClassName] = sSourceValue;
      }
   }
   return fResult;
}

// ***************************************************************************
// results access methods


// ***************************************************************************
//
void
SMDelphi::InitLabels()
{
   m_saLabels.SetSize(TOTAL_COLUMNS);

   // set some default labels and column alignment
   for( int i = 0; i < TOTAL_COLUMNS; ++i )
   {
      CString  sLabel;

      sLabel.LoadString(IDS_DELPHI_DEF_LABEL0 + i);
      m_saLabels[i] = sLabel;
   }
   // now override with user's labels, if any
   dynamic_cast<CSourceMetricsApp *>(AfxGetApp())->
      GetProfileStringList(CSourceMetricsApp::OPTION_SETTINGS + DELPHI_NAME, CSourceMetricsApp::METRICS_LABELS_KEY,
                           m_saLabels, CString(_T(";,")), CString(_T("")));
}

// ***************************************************************************
//
CString
SMDelphi::GetMetricName(int index) const
{
   CString  sMetricName;

   if( (index >= 0) and (index < TOTAL_COLUMNS) )
   {
      sMetricName.LoadString(IDS_DELPHI_METRIC0 + index);
   }
   return sMetricName;
}

// ***************************************************************************
//
CString
SMDelphi::GetValue(int index, bool format /* = true */) const       // access a metric's value
{
   CString  sValue;
   int      iValue = -1;            // start with invalid value
   float    flValue = 0.0;
   int      i;                      // loop index

   sValue.LoadString(IDS_UNDEFINED_VALUE);
   if( (index >= 0) and (m_iCounts.GetSize() > index) )
   {
      switch( index )
      {
      case LINES:
         iValue = m_iCounts[LINE_COUNT];
         break;

      case STATEMENTS:
         iValue = m_iCounts[STATEMENT_COUNT];
         break;

      case PERCENT_COMMENTS:
         if( m_iCounts[LINE_COUNT] > 0 )
         {
            flValue = float(m_iCounts[COMMENT_LINES] * 100) / float(m_iCounts[LINE_COUNT]);
         }
         sValue.Format("%.1f", flValue);
         break;

      case GLOBAL_SUBROUTINES:
         iValue = m_iCounts[GLOBAL_SUBROUTINE_COUNT];
         break;

      case GLOBAL_VARIABLES:
         iValue = m_iCounts[GLOBAL_VARIABLE_COUNT];
         break;

      case LOCAL_SUBROUTINES:
         iValue = m_iCounts[LOCAL_SUBROUTINE_COUNT];
         break;

      case CLASSES:
         iValue = m_iCounts[CLASS_COUNT];
         break;

      case AVERAGE_METHODS_PER_CLASS:
         if( m_iCounts[CLASS_COUNT] > 0 )
         {
            flValue = float(m_iCounts[CLASS_METHOD_COUNT]) / float(m_iCounts[CLASS_COUNT]);
         }
         sValue.Format("%.2f", flValue);
         break;

      case VARIABLES_PER_CLASS:
         if( m_iCounts[CLASS_COUNT] > 0 )
         {
            flValue = float(m_iCounts[CLASS_VARIABLE_COUNT]) / float(m_iCounts[CLASS_COUNT]);
         }
         sValue.Format("%.2f", flValue);
         break;

      case AVERAGE_STATEMENTS_PER_METHOD:
         {
            if( m_iCounts[CLASS_METHOD_COUNT] > 0 )
            {
               flValue = float(m_iCounts[METHOD_STATEMENTS]) / float(m_iCounts[CLASS_METHOD_COUNT]);
            }
            sValue.Format("%.1f", flValue);
         }
         break;

      case BIGGEST_METHOD:
         {
            CString  sClassName;
            CString  sClassValue;
            int      iBiggestLineCount = 0;
            POSITION oListPosition = m_oClassList.GetStartPosition();

            while( oListPosition != NULL )
            {
               m_oClassList.GetNextAssoc(oListPosition, sClassName, sClassValue);
               if( _ttoi(sClassValue) > iBiggestLineCount )
               {
                  iBiggestLineCount = _ttoi(sClassValue);
               }
            }
            if( iBiggestLineCount > 0 )
            {
               iValue = iBiggestLineCount;
            }
            else
            {
               iValue = 0;
            }
         }
         break;

      case AVERAGE_STATEMENTS_PER_SUBROUTINE:
         {
            int   iSubroutines = m_iCounts[GLOBAL_SUBROUTINE_COUNT] + m_iCounts[LOCAL_SUBROUTINE_COUNT];

            if( iSubroutines > 0 )
            {
               flValue = float(m_iCounts[SUBROUTINE_STATEMENTS]) / float(iSubroutines);
            }
            sValue.Format("%.1f", flValue);
         }
         break;

      case BIGGEST_SUBROUTINE:
         {
            if( m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] > 0 )
            {
               iValue = m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS];
            }
            else
            {
               iValue = 0;
            }
         }
         break;

      case PERCENT_BRANCHES:
         if( m_iCounts[STATEMENT_COUNT] > 0 )
         {
            flValue = float(m_iCounts[BRANCH_COUNT] * 100) / float(m_iCounts[STATEMENT_COUNT]);
         }
         sValue.Format("%.1f", flValue);
         break;

      case MAXIMUM_BLOCK_DEPTH:
         {
            sValue.Format("%d +", BLOCK_DEPTH_COUNTS - 1);    // assume the worst

            for( i = 0; i < BLOCK_DEPTH_COUNTS; ++i )
            {
               if( m_iCounts[i + BLOCK_DEPTH_0] == 0 )
               {
                  // counts are base 0, depths are base 0
                  sValue.Format("%ld", max(i - 1, 0));
                  break;
               }
            }
         }
         break;

      case AVERAGE_BLOCK_DEPTH:
         {
            long  lSum = 0;
            long  lCounts = 0;

            for( i = 0; i < BLOCK_DEPTH_COUNTS; ++i )
            {
               long  lCount = m_iCounts[i + BLOCK_DEPTH_0];

               if( lCount > 0 )
               {
                  lSum += lCount * (i + 1);
                  lCounts += lCount;
               }
               else
               {
                  break;
               }
            }
            if( lCounts > 0 )                // if no blocks, avg = 0.0
            {
               flValue = (float(lSum) / float(lCounts)) - (float)1.0;
            }
            sValue.Format("%.2f", flValue);
         }
         break;
      }
   }
   if( iValue >= 0 )
   {
      if( format )
      {
         sValue = FormatCount(iValue);
      }
      else
      {
         sValue.Format("%d", iValue);
      }
   }
   return sValue;
}

// ***************************************************************************
//
bool
SMDelphi::GetDeepestBlockLine(CString & sLabel, int * piResult) const
{
   CString  sDeepestLabel;

   sDeepestLabel.LoadString(IDS_DEEPEST_BLOCK_LABEL);
   sLabel.Format("%s %s)", sDeepestLabel, GetValue(MAXIMUM_BLOCK_DEPTH));
   *piResult = m_iCounts[DEEPEST_BLOCK_LINE];
   return true;
}

// ***************************************************************************
//
bool
SMDelphi::GetBiggestModuleLine(CString & sLabel, int * piResult) const
{
   bool  fResult = false;

   if( (m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] > 0)
   and ((m_iCounts[BIGGEST_METHOD_STATEMENTS] == 0)
     or (m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] > m_iCounts[BIGGEST_METHOD_STATEMENTS])) )
   {
      sLabel.LoadString(IDS_BIGGEST_SUBROUTINE_LABEL);
      *piResult = m_iCounts[BIGGEST_SUBROUTINE_LINE];
      fResult = true;
   }
   else if( (m_iCounts[BIGGEST_METHOD_STATEMENTS] > 0)
   and ((m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] == 0)
      or (m_iCounts[BIGGEST_METHOD_STATEMENTS] >= m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS])) )
   {
      sLabel.LoadString(IDS_BIGGEST_METHOD_LABEL);
      *piResult = m_iCounts[BIGGEST_METHOD_LINE];
      fResult = true;
   }
   return fResult;
}

void
SMDelphi::GetDetails(SMVersion & oVersion, CStringArray & oValues, CDWordArray & oBlockCounts,
                           bool fIncludeExportValues, bool fIncludeEntities, CStringArray & oEntities) const
{
   POSITION oListPosition;
   CString  sClassName;
   CString  sClassValue;
   CString  sBiggestMethod;
   int      iBiggestLineCount;
   CString  sEntity;
   CString  sValue;
   CString  sEmpty;        // used as empty string only

   for( int i = 0; i < TOTAL_COLUMNS; ++i )
   {
      switch( i )
      {
      case MAXIMUM_BLOCK_DEPTH:
         if( oVersion.IsOKVersion(1, 0)                     // added line number of deepest block in V1.1
            and (m_iCounts[DEEPEST_BLOCK_LINE] != UINT_MAX) )
         {
            if( fIncludeEntities )
            {
               sEntity.LoadString(IDS_DEEPEST_LINE_NUMBER);
               oEntities.Add(sEntity);
            }
            sValue.Format("%d", m_iCounts[DEEPEST_BLOCK_LINE]);
            oValues.Add(sValue);
         }
         else if( fIncludeExportValues )
         {
            if( fIncludeEntities )
            {
               sEntity.LoadString(IDS_DEEPEST_LINE_NUMBER);
               oEntities.Add(sEntity);
            }
            oValues.Add(sEmpty);
         }
         break;

      case BIGGEST_METHOD:
         {
            iBiggestLineCount = 0;

            oListPosition = m_oClassList.GetStartPosition();
            while( oListPosition != NULL )
            {
               m_oClassList.GetNextAssoc(oListPosition, sClassName, sClassValue);
               if( _ttoi(sClassValue) > iBiggestLineCount )
               {
                  sBiggestMethod = sClassValue.Mid(sClassValue.Find('@') + 1);
                  iBiggestLineCount = _ttoi(sClassValue);
               }
            }
            if( iBiggestLineCount > 0 )
            {
               sValue.Format("%s.%s(...)", sClassName, sBiggestMethod);
            }
            else
            {
               sValue.LoadString(IDS_DELPHI_NOCLASS_BIGMETHOD);   // set no class message
            }
            if( fIncludeEntities )
            {
               sEntity.LoadString(IDS_DELPHI_BIGGEST_METHOD_NAME);
               oEntities.Add(sEntity);
            }
            oValues.Add(sValue);
            if( (m_iCounts[BIGGEST_METHOD_LINE] != UINT_MAX) and (iBiggestLineCount > 0) )
            {
               if( fIncludeEntities )
               {
                  sEntity.LoadString(IDS_BIGGEST_MODULE_LINE_NUMBER);
                  oEntities.Add(sEntity);
               }
               sValue.Format("%d", m_iCounts[BIGGEST_METHOD_LINE]);
               oValues.Add(sValue);
            }
            else if( fIncludeExportValues )
            {
               if( fIncludeEntities )
               {
                  sEntity.LoadString(IDS_BIGGEST_MODULE_LINE_NUMBER);
                  oEntities.Add(sEntity);
               }
               oValues.Add(sEmpty);
            }
         }
         break;

      case BIGGEST_SUBROUTINE:
         if( fIncludeEntities )
         {
            sEntity.LoadString(IDS_DELPHI_BIGGEST_SUBROUTINE_NAME);
            oEntities.Add(sEntity);
         }
         if( m_sSubroutineName.GetLength() > 0 )
         {
            sValue.Format("%s(...)", m_sSubroutineName);
         }
         else
         {
            sValue.LoadString(IDS_DELPHI_NO_SUBROUTINE_NAME);   // set no subroutine message
         }
         oValues.Add(sValue);
         if( (m_iCounts[BIGGEST_SUBROUTINE_LINE] != UINT_MAX)
         and (m_iCounts[BIGGEST_SUBROUTINE_STATEMENTS] > 0) )
         {
            if( fIncludeEntities )
            {
               sEntity.LoadString(IDS_BIGGEST_SUBROUTINE_LINE_NUMBER);
               oEntities.Add(sEntity);
            }
            sValue.Format("%d", m_iCounts[BIGGEST_SUBROUTINE_LINE]);
            oValues.Add(sValue);
         }
         else if( fIncludeExportValues )
         {
            if( fIncludeEntities )
            {
               sEntity.LoadString(IDS_BIGGEST_SUBROUTINE_LINE_NUMBER);
               oEntities.Add(sEntity);
            }
            oValues.Add(sEmpty);
         }
         break;
      }
      if( fIncludeEntities )
      {
         oEntities.Add(GetMetricName(i));
      }
      oValues.Add(GetValue(i));
   }
   if( not fIncludeExportValues )
   {
      // now display the classes under inserted column labels
      sValue = _T("--------------------------------");
      oEntities.Add(sValue);        // insert separators
      oValues.Add(sValue);
      sEntity = GetMetricName(CLASSES);
      sEntity += _T(":");
      oEntities.Add(sEntity);       // label the items to follow
      sEntity.LoadString(IDS_DELPHI_BIGGEST_METHOD_NAME);
      sValue.Format("%s [%s]:", sEntity, GetMetricName(STATEMENTS));
      oValues.Add(sValue);
      oEntities.Add(sEmpty);        // add blank line
      oValues.Add(sEmpty);
      oListPosition = m_oClassList.GetStartPosition();
      while( oListPosition != NULL )
      {
         m_oClassList.GetNextAssoc(oListPosition, sClassName, sClassValue);
         sBiggestMethod = sClassValue.Mid(sClassValue.Find('@') + 1);
         iBiggestLineCount = _ttoi(sClassValue);
         oEntities.Add(sClassName);
         sValue.Format("%s [%d]", sBiggestMethod, iBiggestLineCount);
         oValues.Add(sValue);
      }
   }
   // now fill the block depth array
   oBlockCounts.SetSize(BLOCK_DEPTH_COUNTS);
   for( int b = 0; b < BLOCK_DEPTH_COUNTS; ++b )
   {
      oBlockCounts[b] = m_iCounts[BLOCK_DEPTH_0 + b];
   }
}

// ***************************************************************************
//
void
SMDelphi::Serialize(CArchive & ar)
{
   // first do the MFC classes and C++ types
	if( ar.IsStoring() )
	{
		ar << m_sSubroutineName;
	}
	else
	{
		ar >> m_sSubroutineName;
	}
   // then the metrics data objects
   SMLanguage::Serialize(ar);
   m_iCounts.Serialize(ar);
   m_oClassList.Serialize(ar);

   return;
}
