/* Copyright (C) 1999, Vsoft Ltd. All Rights Reserved.
 *
 *   Usage Restrictions
 *
 * License is granted to copy, to use, and to make and to use derivative
 * works for research and evaluation purposes only.
 *
 *   Disclaimer of Warranty
 *
 * These software programs are available to the user without any license
 * fee or royalty on an "as is" basis.  Vsoft Ltd. disclaims any and
 * all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any 
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and
 * user's customers, employees, agents, transferees, successors, and
 * assigns.
 *
 * Comments & Questions:
 * mailto:noam@vsoft.com  . Please put "TestPoint" in the subject line.
*/
#include "TestPointManager.h"
#include "CRegistry.h"
#include "VsoftComUtil.h"
#include "stdio.h"

TestPointManager::TestPointManager(IUnknown* pUnknownOuter) 
  : CUnknown(pUnknownOuter)
{
	m_pTpRootPath = "Software\\Vsoft\\TestPoints";
}

TestPointManager::~TestPointManager()
{
	// no need to free m_pTpRootPath since it points to static data.
	m_pTpRootPath = NULL;
}
// check the "IsActive" value in the registry for the subkey idOfTestPoint.
// There is  no default value.
HRESULT TestPointManager::IsActive( const IN GUID & idOfTestPoint)
{
	CRegistry_EX reg;
	DWORD bIsActive;
	LONG r;

	reg.StopParamLogger();
	char szClsid[40], path[1024];
	if(!Guid2Str(idOfTestPoint, szClsid,40))
		return E_FAIL;
	sprintf(path,"%s\\%s",m_pTpRootPath, szClsid);
	r = reg.SetRoot(path);
	if(ERROR_SUCCESS != r)
		return HRESULT_FROM_WIN32(r);
	r = reg.getValueDW("IsActive", &bIsActive);
	if(ERROR_SUCCESS != r)
		return HRESULT_FROM_WIN32(r);
	if(bIsActive)
		return S_OK;
	else
		return S_FALSE;
}


// Get the string from the value "Name" in the registry for the subkey idOfTestPoint
HRESULT TestPointManager::GetNameOfTp( const IN GUID & idOfTestPoint, // ID of the TP
			IN	 UINT bufferSize,		// maximum size of buffer 
			OUT  char* pNameOfTp		// name of the TP will be placed here
			)
{
	CRegistry_EX reg;
	LONG r;
	DWORD stringSize;
	reg.StopParamLogger();
	char szClsid[40], path[1024];
	if(!Guid2Str(idOfTestPoint, szClsid,40))
		return E_FAIL;
	sprintf(path,"%s\\%s",m_pTpRootPath, szClsid);
	r = reg.SetRoot(path);
	if(ERROR_SUCCESS != r)
		return HRESULT_FROM_WIN32(r);

	stringSize = bufferSize; // the 'stringSize' is updated when returning from the next line
	r = reg.getValueREG_SZ("Name",reinterpret_cast<unsigned char*>(pNameOfTp), &stringSize);
	if(ERROR_SUCCESS != r)
		return HRESULT_FROM_WIN32(r);
	return S_OK;
}

// set the "IsActive" flage in the regsitry for the subkey idOfTestPoint
HRESULT TestPointManager::SetActive(const IN GUID & idOfTestPoint , IN bool isActive) 
{
	CRegistry_EX reg;
	LONG r;

	reg.StopParamLogger();
	char szClsid[40], path[1024];
	if(!Guid2Str(idOfTestPoint, szClsid,40))
		return E_FAIL;
	sprintf(path,"%s\\%s",m_pTpRootPath, szClsid);
	r = reg.SetRoot(path);
	if(ERROR_SUCCESS != r)
		return HRESULT_FROM_WIN32(r);

	r = reg.SetValue("IsActive", REG_DWORD, reinterpret_cast<LPBYTE>(&isActive), 4);
	return HRESULT_FROM_WIN32(r);
}

// return to the caller a pointer to the IUnknown interface of the handler
// A handler is created for the given TP, and this handler is initialized by
// calling its ITestPoint::Set() function.
HRESULT TestPointManager::QueryInterfaceForTp(const IN GUID & idOfTestPoint , 
										OUT IUnknown ** ppUnknown)
{
	// find the CLSID of the handler. There is exactly one handler for each TP.

	CRegistry_EX reg;
	LONG r;
	HRESULT hr;
	char clsidString[50];
	CLSID handlerClsid;
	DWORD length = 40;

	reg.StopParamLogger();
	char szClsid[40], path[1024];
	if(!Guid2Str(idOfTestPoint, szClsid,40))
		return E_FAIL;
	sprintf(path,"%s\\%s",m_pTpRootPath, szClsid);
	r = reg.SetRoot(path);
	if(ERROR_SUCCESS != r)
		return HRESULT_FROM_WIN32(r);

	// Look for the handler. ( a COM object )
	// first, look for CLSID value; if it is not found, look for 
	// ProgID.
	// Note: if CLSID entry exists, it overides the ProgId in any case,
	// even if the CLSID is wrong/ill-defined etc.
	r = reg.getValueREG_SZ("CLSID",reinterpret_cast<unsigned char *>(clsidString), &length);
	if( ERROR_FILE_NOT_FOUND == r)
	{	// try "ProgId"
#define MAX_PROGID_NAME 256
		char ProgIdString[MAX_PROGID_NAME];
		r = reg.getValueREG_SZ("ProgID",reinterpret_cast<unsigned char *>(ProgIdString), &length);
		if(ERROR_SUCCESS != r ) 
			return HRESULT_FROM_WIN32(r);
		// convert the char string to wchar_t
		wchar_t wcProgIdString[MAX_PROGID_NAME];
		if(static_cast<UINT>(-1) == mbstowcs( wcProgIdString, ProgIdString, strlen(ProgIdString)+1 )) // convert the null terminator as well
		{
			return E_FAIL;
		}
		if(FAILED(hr = CLSIDFromProgID(wcProgIdString, &handlerClsid))) // beware of this function !
			return hr;
	}
	else
	{
		if(ERROR_SUCCESS != r ) 
			return HRESULT_FROM_WIN32(r);
		if(!Str2Guid(clsidString, handlerClsid))
			return E_FAIL;
	}
	hr = CoCreateInstance(handlerClsid,NULL, CLSCTX_ALL, IID_IUnknown, (void**) ppUnknown);
	if(FAILED(hr))
		return hr;
	// now that we have the handle, give an oportunity to init itself.
	ITestPoint * pStarter;
	hr = (*ppUnknown)->QueryInterface(IID_ITestPoint,(void**) &pStarter);
	if(FAILED(hr))
	{ // the handler is ill behaved!
		(*ppUnknown)->Release();
		return hr;
	}
	hr = pStarter->Set(idOfTestPoint, path);
	if(FAILED(hr))
	{ // tough luck - init of the handler failed.
		pStarter->Release();
		(*ppUnknown)->Release();
		(*ppUnknown) = NULL;
		return hr;
	}
	pStarter->Release();
	return S_OK;
}


// -------------------------------- COM support --------------------------------


/////////////////////////////////////////////////////////// 
// 
// Creation function used by CFactory 
// 
HRESULT TestPointManager::CreateInstance(IUnknown* pUnknownOuter, 
                           CUnknown** ppNewComponent) 
{ 
 *ppNewComponent = new TestPointManager(pUnknownOuter) ; 
 return S_OK ; 
} 


HRESULT  TestPointManager::NondelegatingQueryInterface(const IID& iid, void** ppv) 
{ 
 if (iid == IID_TestPointManager) 
 { 
  return FinishQI(static_cast<ITestPointManager*>(this), ppv) ; 
 } 
 else 
 { 
  return CUnknown::NondelegatingQueryInterface(iid, ppv) ; 
 } 
} 
